from dataclasses import dataclass

import torch
import numpy as np
import random
import pandas as pd
import argparse

from dna_optimizers.reinforce_optimizer import reinforce_optimizer
from dna_optimizers.base_optimizer import get_fitness_info

def get_model_name_or_path(cell):
    if cell == "complex" or cell == "defined":
        return "/blob/ICLR/reglm/yeast_reglm.ckpt"
    elif cell == "hepg2" or cell == "k562" or cell == "sknsh":
        return "/blob/ICLR/reglm/human_reglm.ckpt"
    
def get_prefix_label(task, level):
    
    if task == "complex" or task == "defined":
        if level == "easy":
            return "33"
        elif level == "medium":
            return "22"
        elif level == "hard":
            return "11"
        
    if task == "hepg2":
        if level == "easy":
            return "300"
        elif level == "medium":
            return "200"
        elif level == "hard":
            return "100"
        elif level == 'full':
            return "333"
        
    if task == "k562":
        if level == "easy":
            return "030"
        elif level == "medium":
            return "020"
        elif level == "hard":
            return "010"
        elif level == 'full':
            return "333"
        
    if task == "sknsh":
        if level == "easy":
            return "003"
        elif level == "medium":
            return "002"
        elif level == "hard":
            return "001"
        elif level == 'full':
            return "333"
    else:
        raise ValueError(f"Task {task} not supported.")
    
def set_seed(seed):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
def parse_args():
    parser = argparse.ArgumentParser(description="Parse command line arguments for OptimizerArguments.")

    parser.add_argument("--task", type=str, default="complex", help="The task name.")
    parser.add_argument("--lr", type=float, default=1e-4, help="Learning rate.")
    parser.add_argument("--batch_size", type=int, default=256, help="Batch size.")
    parser.add_argument("--device", type=str, default="cuda", help="Device type (e.g., 'cuda' or 'cpu').")

    parser.add_argument("--max_oracle_calls", type=int, default=4000000, help="Maximum number of oracle calls.")
    parser.add_argument("--max_strings", type=int, default=384000000, help="Maximum number of strings.")
    parser.add_argument("--max_iter", type=int, default=100, help="Maximum number of iterations.")
    parser.add_argument("--epoch", type=int, default=1, help="Number of epochs.")
    
    parser.add_argument("--wandb_log", action='store_true', help="Enable logging with wandb.")
    parser.add_argument("--train_log_interval", type=int, default=4, help="Training log interval.")
    parser.add_argument("--env_log_interval", type=int, default=256, help="Environment logging interval.")

    parser.add_argument("--level", type=str, default="easy", help="Difficulty level.")
    parser.add_argument("--e_size", type=int, default=100, help="Experience size.")
    parser.add_argument("--e_batch_size", type=int, default=24, help="Experience batch size.")
    parser.add_argument("--priority", type=bool, default=True, help="Use priority in experience replay.")

    parser.add_argument("--seed", type=int, default=0, help="Random seed.")
    parser.add_argument("--tfbs_lambda", type=int, default=1, help="Lambda value for TFBS.")
    args = parser.parse_args()
    
    return args
    

def main():
    # optimizer_args = OptimizerArguments()
    optimizer_args = parse_args()
    optimizer_args.prefix_label = get_prefix_label(optimizer_args.task, optimizer_args.level)
    print(f"Prefix Label: {optimizer_args.prefix_label}")
    optimizer_args.model_name_or_path = get_model_name_or_path(optimizer_args.task)
    optimizer_args.max_len, min_fitness, max_fitness = get_fitness_info(optimizer_args.task)
    set_seed(optimizer_args.seed)
    optimizer_args.wandb_run_name = (
        f"{optimizer_args.task}_{optimizer_args.level}_"
        f"e_size_{optimizer_args.e_size}_"
        f"e_bs_{optimizer_args.e_batch_size}_"
        f"lr_{optimizer_args.lr}_"
        f"epoch_{optimizer_args.epoch}_"
        f"prefix_{optimizer_args.prefix_label}_"
        f"seed_{optimizer_args.seed}_"
        "-logp"
)

    print(f"Optimizer Arguments: {optimizer_args}")
    # import debugpy; debugpy.connect(5678); debugpy.wait_for_client();
    optimizer = reinforce_optimizer(optimizer_args)
    
    init_data_path = f"./data/{optimizer_args.task}/{optimizer_args.level}.csv"
    starting_sequences = pd.read_csv(init_data_path)
    
    # min-max normalize the true scores
    starting_sequences['target'] = (starting_sequences['target'] - min_fitness)/(max_fitness - min_fitness)
    starting_sequences = starting_sequences.sort_values('target', ascending=False).head(128)
        
    optimizer.optimize(optimizer_args, starting_sequences)

if __name__ == "__main__":
    main()