import os
import tdc
import itertools
import time
import yaml
import wandb
import torch
import numpy as np

import src.reglm.dataset, src.reglm.lightning, src.reglm.utils, src.reglm.metrics
from .experience import Experience

def evaluate(round_df, starting_sequences):
    data = round_df.sort_values(by='true_score', ascending=False).iloc[:128]
    
    top_fitness = data.iloc[:16]['true_score'].mean().item()
    median_fitness = data['true_score'].median().item()
    
    seqs = data['sequence'].tolist()
    
    distances = [distance(s1, s2) for s1, s2 in itertools.combinations(seqs, 2)]
    diversity = np.median(distances) if distances else 0.0
    
    inits = starting_sequences['sequence'].tolist()
    novelty_distances = [min(distance(seq, init_seq) for init_seq in inits) for seq in seqs]
    novelty = np.median(novelty_distances) if novelty_distances else 0.0
    
    return {
        'top': top_fitness,
        'fitness': median_fitness,
        'diversity': diversity,
        'novelty': novelty
    }
    
    
def get_fitness_info(cell):
    if cell == 'complex':
        length = 80
        min_fitness = 0
        max_fitness = 17
    elif cell == 'defined':
        length = 80
        min_fitness = 0
        max_fitness = 17
    elif cell == 'hepg2':
        length = 200
        min_fitness = -6.051336
        max_fitness = 10.992575
    elif cell == 'k562':
        length = 200
        min_fitness = -5.857445
        max_fitness = 10.781755
    elif cell == 'sknsh':
        length = 200
        min_fitness = -7.283977
        max_fitness = 12.888308
    else:
        raise NotImplementedError()
    return length, min_fitness, max_fitness
    
def top_auc(buffer, top_n, finish, env_log_interval, max_oracle_calls):
    sum = 0
    prev = 0
    called = 0
    ordered_results = list(sorted(buffer.items(), key=lambda kv: kv[1][1], reverse=False))
    for idx in range(env_log_interval, min(len(buffer), max_oracle_calls), env_log_interval):
        temp_result = ordered_results[:idx]
        temp_result = list(sorted(temp_result, key=lambda kv: kv[1][0], reverse=True))[:top_n]
        top_n_now = np.mean([item[1][0] for item in temp_result])
        sum += env_log_interval * (top_n_now + prev) / 2
        prev = top_n_now
        called = idx
    temp_result = list(sorted(ordered_results, key=lambda kv: kv[1][0], reverse=True))[:top_n]
    top_n_now = np.mean([item[1][0] for item in temp_result])
    sum += (len(buffer) - called) * (top_n_now + prev) / 2
    if finish and len(buffer) < max_oracle_calls:
        sum += (max_oracle_calls - len(buffer)) * top_n_now
    return sum / max_oracle_calls

def distance(s1, s2):
    return sum([1 if i != j else 0 for i, j in zip(list(s1), list(s2))])

def diversity(seqs):
    divs = []
    for s1, s2 in itertools.combinations(seqs, 2):
        divs.append(distance(s1, s2))
    return sum(divs) / len(divs)

def mean_distance(seq, seqs):
    divs = []
    for s in seqs:
        divs.append(distance(seq, s))
    return sum(divs) / len(divs)

class BaseOptimizer:
    def __init__(self, cfg=None):
        self.cfg = cfg
        
        self.task = cfg.task
        _, self.min_fitness, self.max_fitness = get_fitness_info(self.task)
        self.assign_target(cfg)

        self.max_oracle_calls = cfg.max_oracle_calls
        self.env_log_interval = cfg.env_log_interval
        
        self.dna_buffer = dict()
        self.mean_score = 0

        #logging counters
        self.last_log = 0
        self.last_log_time = time.time()
        self.last_logging_time = time.time()
        self.total_count = 0
        self.invalid_count = 0
        self.redundant_count = 0
        
        if cfg.wandb_log:
            wandb.init(
                # entity="bamal",
                project="Hyena_RL",
                # dir=cfg.wandb_dir,
                name=cfg.wandb_run_name,
            )
            self.define_wandb_metrics()
        #set device
        self.device = torch.device(cfg.device)
        # replay buffer        
        self.experience = Experience(cfg.e_size, cfg.priority)

    @property
    def budget(self):
        return self.max_oracle_calls
    
    @property
    def finish(self):
        return len(self.dna_buffer) >= self.max_oracle_calls
    
    def assign_target(self, cfg): 
        # Check if task is yeast promoter
        if self.task == "complex":
            # Load complext_target from checkpoint
            self.target = src.reglm.regression.EnformerModel.load_from_checkpoint(
                "/home/aiscuser/data/yeast_regression_paired_complex.ckpt",
            ).to(cfg.device)
        elif self.task == "defined":
            self.target = src.reglm.regression.EnformerModel.load_from_checkpoint(
                "/home/aiscuser/data/yeast_regression_paired_defined.ckpt",
            ).to(cfg.device)
            
        elif self.task == 'hepg2':
            self.target = src.reglm.regression.EnformerModel.load_from_checkpoint(
                "/home/aiscuser/data/human_regression_paired_hepg2.ckpt",
            ).to(cfg.device)
        elif self.task == 'k562':
            self.target = src.reglm.regression.EnformerModel.load_from_checkpoint(
                "/home/aiscuser/data/human_regression_paired_k562.ckpt",
            ).to(cfg.device)
            
        elif self.task == 'sknsh':
            self.target = src.reglm.regression.EnformerModel.load_from_checkpoint(
                "/home/aiscuser/data/human_regression_paired_sknsh.ckpt",
            ).to(cfg.device)
        else:
            raise NotImplementedError
        
        self.target.eval()            

            
    
    def define_wandb_metrics(self):
        #new wandb metric
        wandb.define_metric("num_dnas")
        wandb.define_metric("avg_top1", step_metric="num_dnas")
        wandb.define_metric("avg_top10", step_metric="num_dnas")
        wandb.define_metric("avg_top100", step_metric="num_dnas")
        wandb.define_metric("auc_top1", step_metric="num_dnas")
        wandb.define_metric("auc_top10", step_metric="num_dnas")
        wandb.define_metric("auc_top100", step_metric="num_dnas")
        wandb.define_metric("avg_sa", step_metric="num_dnas")
        wandb.define_metric("diversity_top100", step_metric="num_dnas")
        wandb.define_metric("n_oracle", step_metric="num_dnas")
        wandb.define_metric("invalid_count", step_metric="num_dnas")
        wandb.define_metric("redundant_count", step_metric="num_dnas")

    def normalize_target(self, score):
        return (score - self.min_fitness) / (self.max_fitness - self.min_fitness)
    
    @torch.no_grad()
    def score_enformer(self, dna):
        if len(self.dna_buffer) > self.max_oracle_calls:
            return 0
        
        score = self.target([dna]).squeeze(0).item()
        score = self.normalize_target(score)
        # score = self.oracle([dna]).squeeze(0).item()
        if dna in self.dna_buffer:
            self.dna_buffer[dna][2] += 1
            self.redundant_count += 1

        else:
            self.dna_buffer[dna] = [float(score), len(self.dna_buffer)+1, 1]

        return self.dna_buffer[dna][0]
    
    def add_dna(self, dna, score):
        if dna in self.dna_buffer:
            self.dna_buffer[dna][2] += 1
            self.redundant_count += 1
        else:
            self.dna_buffer[dna] = [float(score), len(self.dna_buffer)+1, 1]

    def predict_enformer(self, dna_list):
        st = time.time()
        assert type(dna_list) == list
        self.total_count += len(dna_list)
        score_list = []
        for dna in dna_list:
            score_list.append(self.score_enformer(dna))
                
            if len(self.dna_buffer) % self.env_log_interval == 0 and len(self.dna_buffer) > self.last_log:
                self.sort_buffer()
                self.log_intermediate()
                self.last_log_time = time.time()
                self.last_log = len(self.dna_buffer)

        self.last_logging_time = time.time() - st
        self.mean_score = np.mean(score_list)
        return score_list

    def optimize(self, cfg):
        raise NotImplementedError

    def sort_buffer(self):
        self.dna_buffer = dict(sorted(self.dna_buffer.items(), key=lambda kv: kv[1][0], reverse=True))
            
    def log_intermediate(self, dna=None, scores=None, finish=False):
        if finish:
            temp_top100 = list(self.dna_buffer.items())[:100]
            dnas = [item[0] for item in temp_top100]
            scores = [item[1][0] for item in temp_top100]
            n_calls = self.max_oracle_calls
        
        else:
            if dna is None and scores is None:
                if len(self.dna_buffer) <= self.max_oracle_calls:
                    # If not spefcified, log current top-100 dna in buffer
                    temp_top100 = list(self.dna_buffer.items())[:100]
                    dnas = [item[0] for item in temp_top100]
                    scores = [item[1][0] for item in temp_top100]
                    n_calls = len(self.dna_buffer)
                else:
                    results = list(sorted(self.dna_buffer.items(), key=lambda kv: kv[1][1], reverse=False))[:self.max_oracle_calls]
                    temp_top100 = sorted(results, key=lambda kv: kv[1][0], reverse=True)[:100]
                    dnas = [item[0] for item in temp_top100]
                    scores = [item[1][0] for item in temp_top100]
                    n_calls = self.max_oracle_calls
            else:
                raise NotImplementedError
       
        avg_top1 = np.max(scores)
        avg_top10 = np.mean(sorted(scores, reverse=True)[:10])
        avg_top100 = np.mean(scores)

    
        # diversity_top100 = diversity(dnas)
        # print(f'{n_calls}/{self.max_oracle_calls} | '
        #         f'avg_top1: {avg_top1:.3f} | '
        #         f'avg_top10: {avg_top10:.3f} | '
        #         f'avg_top100: {avg_top100:.3f} | '
        #         f'time: {time.time() - self.last_log_time:.3f} | '
        #         f'logging time : {self.last_logging_time} | '
        #         f'mean_score: {self.mean_score:.3f} | '
        #         f'tot_cnt: {self.total_count} | '
        #         f'red_cnt: {self.redundant_count} | '
        #         f'div_top100: {diversity_top100:.3f}'
        #         )

        # if self.cfg.wandb_log: 
        #     wandb.log({
        #         "avg_top1": avg_top1, 
        #         "avg_top10": avg_top10, 
        #         "avg_top100": avg_top100,
        #         "auc_top1": top_auc(self.dna_buffer, 1, finish, self.env_log_interval, self.max_oracle_calls),
        #         "auc_top10": top_auc(self.dna_buffer, 10, finish, self.env_log_interval, self.max_oracle_calls),
        #         "auc_top100": top_auc(self.dna_buffer, 100, finish, self.env_log_interval, self.max_oracle_calls),
        #         "diversity_top100": diversity_top100,
        #         "invalid_count" : self.invalid_count,
        #         "redundant_count": self.redundant_count,
        #         "num_dnas": n_calls,
        #     })

           
            # data = [[scores[i], docking_scores[i], smis[i], wandb.Image(Draw.MolToImage(Chem.MolFromSmiles(smis[i])))] for i in range(10)]

            # columns = ["Score", "Docking score", "SMILES", "IMAGE"]
            # wandb.log({"Top 10 dnas": wandb.Table(data=data, columns=columns)})
    
    def _analyze_results(self, results):
        results = results[:100]
        scores_dict = {item[0]: item[1][0] for item in results}
        smis = [item[0] for item in results]
        scores = [item[1][0] for item in results]
        smis_pass = self.filter(smis)
        if len(smis_pass) == 0:
            top1_pass = -1
        else:
            top1_pass = np.max([scores_dict[s] for s in smis_pass])
        return [np.mean(scores), 
                np.mean(scores[:10]), 
                np.max(scores), 
                self.diversity_evaluator(smis), 
                np.mean(self.sa_scorer(smis)), 
                float(len(smis_pass) / 100), 
                top1_pass]
    
    def save_result(self, suffix=None):
        if suffix is None:
            output_file_path = os.path.join(self.cfg.output_dir, 'results.yaml')
        else:
            output_file_path = os.path.join(self.cfg.output_dir, 'results_' + suffix + '.yaml')
        
        self.sort_buffer()
        with open(output_file_path, 'w') as f:
            yaml.dump(self.dna_buffer, f, sort_keys=False)

    def __len__(self):
        return len(self.dna_buffer)