import os
import sys
import wandb
import hydra
import torch
from torch import optim
import numpy as np
import json
import pandas as pd

import src.reglm.regression

from .base_optimizer import BaseOptimizer
from .reinforce_optimizer import reinforce_optimizer
import src.reglm.dataset, src.reglm.lightning, src.reglm.utils, src.reglm.metrics
import scripts.utils, scripts.motifs

from dna_optimizers.base_optimizer import evaluate

def get_params(model):
    return (p for p in model.parameters() if p.requires_grad)

class tfbs_optimizer(reinforce_optimizer):
    def __init__(self, cfg=None):
        super().__init__(cfg)
        
        self.tfbs_lambda = cfg.tfbs_lambda
        motifs, bg = scripts.motifs.read_meme(
            cfg.meme_path
        )
        self.tfbs_reward_dict = json.load(open(cfg.tfbs_reward_path, 'r'))
        sel = scripts.utils.load_csv(
            cfg.ppms_path
        ).Matrix_id.tolist()
        self.motif2idx = {name: i for i, name in enumerate(sel)}
        self.idx2motif = {i: name for i, name in enumerate(sel)}
        motifs = [m for m in motifs if m.name.decode() in sel]
        self.motifs = motifs
        self.bg = bg


    def optimize(self, cfg, starting_sequences):

        # results' dataframe
        columns = ['round', 'sequence', 'true_score']
        df = pd.DataFrame(columns=columns)
        summary_columns = ["round", "top", "fitness", "diversity", "novelty"]
        summary_df = pd.DataFrame(columns=summary_columns)
        
        # 1 add inital experience
        init_seqs = starting_sequences['sequence'].tolist()
        init_tokens = self.agent.encode(starting_sequences['sequence'].tolist(), [self.prefix_label] * len(init_seqs), add_start=True)
        init_tokens = init_tokens.T.to(self.device)
        # 
        init_rewards = torch.zeros(cfg.max_len+len(self.prefix_label), len(init_seqs))
        init_scores = starting_sequences['target'].tolist()
        init_rewards[-1, :] = torch.tensor(init_scores, dtype=torch.float32)
        init_rewards = init_rewards.to(self.device)
        # init_scores = torch.tensor(init_scores, dtype=torch.float32).to(self.device)
        init_nonterms = [False] * len(self.prefix_label) + [True] * (cfg.max_len) + [False] 
        init_nonterms = torch.tensor(init_nonterms, dtype=torch.bool)
        init_nonterms = init_nonterms.unsqueeze(-1).expand(-1, len(init_seqs)).to(self.device)
        init_lens = [cfg.max_len] * len(init_seqs)
        init_lens = torch.tensor(init_lens, dtype=torch.long)
        self.experience.add_experience(init_seqs, init_tokens, init_scores, init_rewards, init_nonterms, init_lens)
        
        for i in range(cfg.epoch):
            self.update(init_tokens, init_rewards, init_nonterms, init_lens, cfg, dict(), False)
        
        train_steps = 0
        eval_strings = 0
        metrics = dict() 
        print('Start training ... ')
        # while eval_strings < cfg.max_strings:
        for it in range(1, cfg.max_iter + 1):

            with torch.no_grad():
                # sample experience
                # labels = ["44"] * cfg.batch_size
                labels = [self.prefix_label] * cfg.batch_size
                obs, rewards, nonterms, episode_lens = self.agent.get_data(labels, cfg.max_len)

            dna_list = []
            for dna in obs.cpu().numpy().T:
                dna_seq = self.agent.decode(dna, ignore_num=len(self.prefix_label)+1)[0]
                assert len(dna_seq) == cfg.max_len
                dna_list.append(dna_seq)
                
            score = np.array(self.predict(dna_list))
            scores = torch.tensor(score, dtype=torch.float32, device=self.device)
             
             
            # ********* TFBS **********
            # import pdb; pdb.set_trace()
            tfbs_sites = scripts.motifs.scan(dna_list, self.motifs, self.bg)
            tfbs_sites["start"] -= 1 # 1-based to 0-based
            tfbs_struc = scripts.motifs.generate_tfbs_structure(tfbs_sites)
            for idx, tfbs_info in enumerate(tfbs_struc):
                seq_id = list(tfbs_info.keys())[0]
                tfbs_list = tfbs_info[seq_id]
                for tfbs in tfbs_list:
                    start, end = tfbs['start'], tfbs['end']
                    region_reward = self.tfbs_reward_dict[tfbs['Matrix_id']] * self.tfbs_lambda
                    rewards[end+len(self.prefix_label)-1, idx] += region_reward # end+2-1 because of 1-based indexing

            # for idx, tfbs_info in enumerate(tfbs_struc):
            #     seq_id = list(tfbs_info.keys())[0]
            #     tfbs_list = tfbs_info[seq_id]

            #     max_reward_tfbs = None
            #     max_abs_reward = float('-inf')
                
            #     for tfbs in tfbs_list:
            #         region_reward = abs(self.tfbs_reward_dict[tfbs['Matrix_id']] * self.tfbs_lambda)
            #         if region_reward > max_abs_reward:
            #             max_abs_reward = region_reward
            #             max_reward_tfbs = tfbs
                

            #     if max_reward_tfbs is not None:
            #         start, end = max_reward_tfbs['start'], max_reward_tfbs['end']
            #         region_reward = self.tfbs_reward_dict[max_reward_tfbs['Matrix_id']] * self.tfbs_lambda
            #         rewards[end + len(self.prefix_label) - 1, idx] += region_reward
            # for idx, tfbs_info in enumerate(tfbs_struc):
            #     seq_id = list(tfbs_info.keys())[0]
            #     tfbs_list = tfbs_info[seq_id]

            #     for tfbs in tfbs_list:
            #         start, end = tfbs['start'], tfbs['end']
            #         region_reward = self.tfbs_reward_dict[tfbs['Matrix_id']] * self.tfbs_lambda

            #         # 仅对负的 reward 进行更新
            #         if region_reward < 0:
            #             rewards[end + len(self.prefix_label) - 1, idx] += region_reward

            # ********* TFBS **********

            train_steps += 1
            # eval_strings += cfg.batch_size

            log = False
            if cfg.wandb_log and train_steps % cfg.train_log_interval == 0:
                log = True
                metrics = dict()
                metrics['eval_strings'] = eval_strings
                metrics['mean_score'] = np.mean(score)
                metrics['max_score'] = np.max(score)
                metrics['min_score'] = np.min(score)
                metrics['mean_episode_lens'] = np.mean(episode_lens.tolist())
                metrics['max_episode_lens'] = np.max(episode_lens.tolist())
                metrics['min_episode_lens'] = np.min(episode_lens.tolist())
                wandb.log(metrics)

            rewards[-1, :] += scores
            # replay buffer
            if len(self.experience) > cfg.e_batch_size:
                e_obs, e_scores, e_rewards, e_nonterms, e_episode_lens = self.experience.sample(cfg.e_batch_size, self.device)
                e_L, e_B = e_obs.shape
                L, B = obs.shape

                f_L = max(e_L, L)

                f_obs = torch.zeros((f_L, cfg.batch_size + cfg.e_batch_size), dtype=torch.long, device=self.device)
                f_nonterms = torch.zeros((f_L, cfg.batch_size + cfg.e_batch_size), dtype=torch.bool, device=self.device)

                f_obs[:L, :B] = obs
                f_obs[:e_L, B:] = e_obs

                f_nonterms[:L, :B] = nonterms
                f_nonterms[:e_L, B:] = e_nonterms
                
                # f_scores = torch.cat([scores, e_scores], dim=-1)
                f_rewards = torch.cat([rewards, e_rewards], dim=-1)
                f_episode_lens = torch.cat([episode_lens, e_episode_lens])
               
                for i in range(cfg.epoch):
                    self.update(f_obs, f_rewards, f_nonterms, f_episode_lens, cfg, metrics, log)
            else:
                for i in range(cfg.epoch):
                    # import pdb; pdb.set_trace()
                    self.update(obs, rewards, nonterms, episode_lens, cfg, metrics, log)
                
            self.experience.add_experience(dna_list, obs, score, rewards, nonterms, episode_lens)

            round_df = pd.DataFrame({'round': [it]*len(dna_list), 'sequence': dna_list, 'true_score': score})
            df = pd.concat([df, round_df], ignore_index=True)
            
            round_results = evaluate(round_df, starting_sequences)
            
            round_results['round'] = it
            
            round_results_df = pd.DataFrame([round_results])
            summary_df = pd.concat([summary_df, round_results_df], ignore_index=True)
            
            if cfg.wandb_log:
                wandb.log(round_results)
            print(f"Round {it} finished")
            print(round_results)
            
        # save results
        # ./results/{cfg.task}_{cfg.level}_{cfg.seed}.csv
        os.makedirs('results', exist_ok=True)
        save_name = f'results/tfbs_{cfg.tfbs_lambda}_{cfg.task}_{cfg.level}_{cfg.seed}.csv'
        df.to_csv(save_name, index=False)
        
        summary_name = f'results/tfbs_{cfg.tfbs_lambda}_{cfg.task}_{cfg.level}_{cfg.seed}_summary.csv'
        summary_df.to_csv(summary_name, index=False)
        
        print('max training string hit')
        wandb.finish()
        sys.exit(0)
    # def optimize(self, cfg):

    #     train_steps = 0
    #     eval_strings = 0
    #     metrics = dict() 
    #     print('Start training ... ')
    #     while eval_strings < cfg.max_strings:

    #         with torch.no_grad():
    #             labels = [self.prefix_label] * cfg.batch_size
    #             obs, rewards, nonterms, episode_lens = self.agent.get_data(labels, cfg.max_len)

    #         dna_list = []
    #         for dna in obs.cpu().numpy().T:
    #             dna_seq = self.agent.decode(dna, ignore_num=len(self.prefix_label)+1)[0]
    #             dna_list.append(dna_seq)
                
    #         score = np.array(self.predict(dna_list))
    #         scores = torch.tensor(score, dtype=torch.float32, device=self.device)
             
    #         if self.finish:
    #             print('max oracle hit')
    #             wandb.finish()
    #             sys.exit(0)


    #         # ********* TFBS **********
    #         tfbs_sites = scripts.motifs.scan(dna_list, self.motifs, self.bg)
    #         tfbs_sites["start"] -= 1 # 1-based to 0-based
    #         tfbs_struc = scripts.motifs.generate_tfbs_structure(tfbs_sites)
    #         for idx, tfbs_info in enumerate(tfbs_struc):
    #             seq_id = list(tfbs_info.keys())[0]
    #             tfbs_list = tfbs_info[seq_id]
    #             for tfbs in tfbs_list:
    #                 start, end = tfbs['start'], tfbs['end']
    #                 region_reward = self.tfbs_reward_dict[tfbs['Matrix_id']] * self.tfbs_lambda
    #                 rewards[end+len(self.prefix_label)-1, idx] += region_reward # end+2-1 because of 1-based indexing
    #         # ********* TFBS **********
                      
    #         train_steps += 1
    #         eval_strings += cfg.batch_size
            
    #         log = False
    #         if cfg.wandb_log and train_steps % cfg.train_log_interval == 0:
    #             log = True
    #             metrics = dict()
    #             metrics['eval_strings'] = eval_strings
    #             metrics['mean_score'] = np.mean(score)
    #             metrics['max_score'] = np.max(score)
    #             metrics['min_score'] = np.min(score)
    #             metrics['mean_episode_lens'] = np.mean(episode_lens.tolist())
    #             metrics['max_episode_lens'] = np.max(episode_lens.tolist())
    #             metrics['min_episode_lens'] = np.min(episode_lens.tolist())
    #             wandb.log(metrics)

    #         rewards[-1, :] += scores
    #         # replay buffer
    #         if len(self.experience) > cfg.batch_size:
    #             e_obs, e_scores, e_nonterms, e_episode_lens = self.experience.sample(cfg.e_batch_size, self.device)
    #             e_L, e_B = e_obs.shape
    #             L, B = obs.shape

    #             f_L = max(e_L, L)

    #             f_obs = torch.zeros((f_L, cfg.batch_size + cfg.e_batch_size), dtype=torch.long, device=self.device)
    #             f_nonterms = torch.zeros((f_L, cfg.batch_size + cfg.e_batch_size), dtype=torch.bool, device=self.device)

    #             f_obs[:L, :B] = obs
    #             f_obs[:e_L, B:] = e_obs

    #             f_nonterms[:L, :B] = nonterms
    #             f_nonterms[:e_L, B:] = e_nonterms

    #             f_scores = torch.cat([scores, e_scores], dim=-1)
    #             f_episode_lens = torch.cat([episode_lens, e_episode_lens])
               
    #             self.update(f_obs, f_scores, f_nonterms, f_episode_lens, cfg, metrics, log)
    #         else:
    #             self.update(obs, rewards, nonterms, episode_lens, cfg, metrics, log)
                
    #         self.experience.add_experience(dna_list, obs, score, nonterms, episode_lens)

    #     print('max training string hit')
    #     wandb.finish()
    #     sys.exit(0)